
package fs;

import ev.Evolution;
import ev.Geno;
import ev.Pheno;
import fs.data.DataSourceAS;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;
import java.util.logging.Logger;
import weka.classifiers.Classifier;

/**
 *
 * @author sigved
 */
public class FeatureSet extends Pheno implements Geno{
    protected final Random r;
    
    protected final int numberOfFeatures;
    protected final ArrayList<Class<? extends Classifier>> availableClassifiers;
    
    public Set<Integer> indexes = new HashSet<Integer>();
    public Class<? extends Classifier> classifier;
    
    /**
     * internal constructor that creates a empty feature set without a specified classifier, to get randomized featureSets call createFeaturePopulation or createRandomFeatureSet
     * @param nfs number of features in set it's to be used on
     * @param availableClassifiers a set of possible classifiers to use
     */
    protected FeatureSet(long seed, int nfs, ArrayList<Class<? extends Classifier>> availableClassifiers){
        this.r = new Random(seed);
        this.numberOfFeatures = nfs;
        this.availableClassifiers = availableClassifiers;
        this.setFitness(-1d);
    }

        
    @Override
    public Pheno createPheno() {
        return this;
    }
    
    @Override
    public Geno getGeno() {
        return this;
    }
    
    @Override
    public Geno combineWith(Geno... g) {
//        System.out.println(g.length);
        //check for incompatible genos
        for(Geno geno : g)if(!(geno instanceof FeatureSet)){
            Logger.getLogger(FeatureSet.class.getName()).severe("Combining featureSet geno with something that isn't");
            return null;
        }
        
        //create a list of all the genos indexes, list not set so that popular indexes are more likely to be included
        ArrayList<Integer> newindexes = new ArrayList<Integer>(this.indexes);
        for(Geno geno : g)newindexes.addAll(((FeatureSet)geno).indexes);
        
        //determine the size of the new pheno
        double size = this.indexes.size();
        for(Geno geno : g)size += ((FeatureSet)geno).indexes.size();
        size = size/(1.0d+(double)g.length);
        if(size%1d > 0.001d){
            if(r.nextBoolean())size = Math.floor(size);
            else size = Math.ceil(size);
        }
        
        //create new featureSet and select indexes
        FeatureSet fs = new FeatureSet(this.r.nextLong(),numberOfFeatures, availableClassifiers);
        while(fs.indexes.size() < Math.round(size) && newindexes.size() > 0){
            fs.indexes.add(newindexes.remove(r.nextInt(newindexes.size())));
        }
        
        //pick a classifier type from a random parent
        int nclassifier = r.nextInt(g.length+1);
        if(nclassifier < g.length)fs.classifier = ((FeatureSet)g[nclassifier]).classifier;
        else fs.classifier = this.classifier;
        
        return fs;
    }

    @Override
    public void mutate(double rate) {
        int size = indexes.size();
        ArrayList<Integer> pick = new ArrayList<Integer>(indexes);
        //randomly remove or add some index
        if(r.nextFloat() < rate){
            if(r.nextFloat() > 0.5f && indexes.size() > 0){
                indexes.remove(pick.get(r.nextInt(size)));
            }else{
                indexes.add(r.nextInt(numberOfFeatures));
            }
        }
        
//        if(indexes.size() > 0 && r.nextFloat() < rate)indexes.remove(pick.get(r.nextInt(size)));
//        
//        //randomly add some index
//        if(r.nextFloat() < rate)indexes.add(r.nextInt(numberOfFeatures));
        
        //randomly change the classifier
        if(r.nextDouble() < rate)classifier = availableClassifiers.get(r.nextInt(availableClassifiers.size()));
    }
    
    @Override
    public String toString(){
        return this.classifier.toString() + indexes.toString() + "age(" + this.getAge() + ") " + ((float)this.getFitness());
    }
    
    /**
     * @param numberOf phenos in population
     * @param classifiers a list of possible classifers to use
     * @param numberOfFeatures
     * @return a population of featureSets built as specified
     */
    public static Evolution<FeatureSet, FeatureSet> createFeaturePopulation(long seed, int numberOf, ArrayList<Class<? extends Classifier>> classifiers, int numberOfFeatures, boolean empty){
        Evolution<FeatureSet,FeatureSet> ret = new Evolution<FeatureSet, FeatureSet>();
        Random nr = new Random(seed);
        
        for(int i = 0; i < numberOf; i++)ret.addPheno(createRandomFeatureSet(nr.nextLong(), numberOfFeatures, classifiers, empty));
        
        return ret;
    }
    
    /**
     * @param numberOfFeatures
     * @param classifiers
     * @return a pseudo random feature selection, given available features and classifiers
     */
    public static FeatureSet createRandomFeatureSet(long seed, int numberOfFeatures, ArrayList<Class<? extends Classifier>> classifiers, boolean empty){
        Random nr = new Random(seed);
        
        FeatureSet fs = new FeatureSet(nr.nextInt(),numberOfFeatures, classifiers);
        
        if(!empty)for(int i = 0; i < 15; i++)if(nr.nextBoolean())fs.indexes.add(nr.nextInt(numberOfFeatures));
        
        fs.classifier = classifiers.get(nr.nextInt(classifiers.size()));
        
        return fs;
    }
    
    public static TreeSet<FeatureSet> allPossibleFeatureSetsSorted(Class<? extends Classifier> c, DataSourceAS ds){
        WrapperFitness wf = new WrapperFitness(ds);
        Set<FeatureSet> fss = new HashSet<FeatureSet>();
        int na = ds.getAll().numAttributes();
        Evolution<FeatureSet, FeatureSet> e = new Evolution<FeatureSet, FeatureSet>();
        for(Set<Integer> ints: Test.getAllPossibleSubsets(na)){
            FeatureSet fs = new FeatureSet(0,na, null);
            fs.indexes = ints;
            fs.classifier = c;
            fss.add(fs);
            e.addPheno(fs);
        }
        wf.runOnPopulation(fss);
        return e.sortedByFitness();
    }
    
    final public static int EMPTY = 0;
    final public static int ONE = 1;
    final public static int RANDOM = 2;
    final public static int EACH = 3;
    static void initiatePopulation(int initpop, Set<FeatureSet> pop, long seed, WrapperFitness wf) throws Exception {
        Random nr = new Random(seed);
        if(initpop == EMPTY){
            for(FeatureSet fs : pop)fs.indexes = new HashSet<Integer>();
        }else if(initpop == ONE){
            for(FeatureSet fs : pop)fs.indexes = new HashSet<Integer>();
            for(FeatureSet fs : pop)fs.indexes.add(nr.nextInt(fs.numberOfFeatures));
        }else if(initpop == RANDOM){
            for(FeatureSet fs : pop){
                fs.indexes = new HashSet<Integer>();
                int nf = nr.nextInt(fs.numberOfFeatures+1);
                ArrayList<Integer> addfrom = new ArrayList<Integer>();
                for(int i = 0; i < fs.numberOfFeatures; i++)addfrom.add(i);
                for(int i = 0; i < nf && i < 100; i++){
                    fs.indexes.add(addfrom.remove(nr.nextInt(addfrom.size())));
                }
            }
        }else if(initpop == EACH){
            FeatureSet orig = pop.iterator().next();
            pop.clear();
            for(int i = 0; i < orig.numberOfFeatures; i++){
                FeatureSet nfs = new FeatureSet(nr.nextLong(), orig.numberOfFeatures, orig.availableClassifiers);
                nfs.indexes = new HashSet<Integer>();
                nfs.indexes.add(i);
                nfs.classifier = orig.classifier;
                pop.add(nfs);
            }
            wf.runOnPopulation(pop);
        }else{
            throw new Exception("WRONG INITIAL POPULATION SETTINGS");
        }
    }
}